/* Copyright 2012 Perttu Ahola / celeron55 */

var NumMaps = 4;

console.log("Note: Required modules: socket.io, mime");

var helpers = require("./nodescripts/helpers.js");
var dump = helpers.dump;

var app = require('http').createServer(handler)
	, io = require('socket.io').listen(app)
	, fs = require('fs')

app.listen(8082);

function handler(req, res){
	filemap = [];
	filemap["/"] = "suunnistus.html";
	filename = filemap[req.url];
	if(filename === undefined){
		/*console.log("handler: uknown url: "+req.url);
		return;*/
		/* Very unsafe! */
		filename = req.url;
		if(filename.indexOf("..") != -1){
			console.log("Disallowed "+filename);
			res.writeHead(500);
			return res.end('Error loading '+filename);
		}
	}
	
	console.log("handler: serving url: "+req.url);
	
	var path = __dirname + '/' + filename;
	type = require('mime').lookup(path);
	console.log("type="+type);
	fs.readFile(path,
	function (err, data) {
		if (err) {
			res.writeHead(500);
			return res.end('Error loading '+filename);
		}
		res.writeHead(200, {'Content-Type': type});
		res.end(data);
	});
}

/*function send_world(socket, world){
	socket.emit("world", {
		w: world.w,
		h: world.h,
		world: world.serialize()
	});
}*/

// key = client_uid, value = socket
var clients = [];
var clients_in_game = [];

function get_client_uid(socket){
	for(var uid in clients){
		client = clients[uid];
		if(client == socket){
			return uid;
		}
	}
	return undefined;
}

// key = client_uid, value = name
var names = [];

function GameClient(uid){
	this.uid = uid;
}

function Player(client_uid){
	this.client_uid = client_uid;
	this.path = undefined;
	this.ended = false;
}

function Game(){
	this.ended = false; // When game has ended, this is set to true
	this.clients = {}; // GameClient objects, indexed by uid
	this.players = {}; // Player objects, indexed by uid
	
	this.map_spec = {};
	this.map_spec.index = Math.floor(Math.random()*NumMaps);
	this.map_spec.home = {};
	this.map_spec.home.x = Math.random()*640;
	this.map_spec.home.y = Math.random()*480;
	this.map_spec.checkpoints = [
		{x: Math.random()*640, y:Math.random()*480},
		{x: Math.random()*640, y:Math.random()*480},
		{x: Math.random()*640, y:Math.random()*480}
	];

	this.timeout_timer = 120;
	this.timeout_triggered = false;

	this.emit_to_all = function(cmd, stuff){
		for(var i in this.clients){
			var uid = this.clients[i].uid;
			var socket = clients[uid];
			if(socket)
				socket.emit(cmd, stuff);
		}
	}

	this.get_player_names = function(){
		var names = [];
		for(var i in this.players){
			names.push(this.players[i].name);
		}
		return names;
	}

	this.add_player = function(uid){
		//var socket = clients[uid];
		this.clients[uid] = new GameClient(uid);
		this.players[uid] = new Player(uid);
	}

	this.all_paths_set = function(){
		for(var i in this.players){
			var player = this.players[i];
			if(player.ended == true)
				continue;
			if(player.path == undefined)
				return false;
		}
		return true;
	}

	this.all_ended = function(){
		for(var i in this.players){
			var player = this.players[i];
			if(player.ended == false)
				return false;
		}
		return true;
	}

	this.emit_startup_info = function(){
		var stuff = {
			map_spec: this.map_spec,
			players: {},
		};
		for(var uid in this.players){
			var player = this.players[uid];
			stuff.players[uid] = {
				name: names[uid],
			};
		}
		this.emit_to_all('startup_info', stuff);
	}

	this.emit_all_paths = function(){
		var stuff = {
			all_paths: {},
		};
		console.log("emitting all paths:");
		for(var uid in this.players){
			var player = this.players[uid];
			stuff.all_paths[uid] = {
				path: player.path,
			};
		}
		this.emit_to_all('all_paths', stuff);
	}

	this.kick_players_with_no_path_and_start = function(){
		var any_kicked = false;
		for(var uid in this.players){
			var player = this.players[uid];
			if(player.path == undefined){
				var socket = clients[uid];
				if(socket)
					socket.emit('kick', {});
				delete this.players[uid];
				delete this.clients[uid];
				clients_in_game[uid] = false;
				any_kicked = true;
			}
		}
		if(any_kicked){
			this.emit_all_paths();
		}
	}

	this.tick = function(){
		this.timeout_timer -= 1;
		if(this.timeout_timer < 0 && !this.timeout_triggered){
			this.timeout_triggered = true;
			this.kick_players_with_no_path_and_start();
		}
		
		this.send_status();
		
		if(!this.ended){
			var game = this
			setTimeout(function(){
				game.tick();
			}, 1000);
		}
	}

	this.send_status = function(){
		this.emit_to_all('status', {
			timeout_timer: this.timeout_timer,
		});
	}
}

var games = []; // A list of games

function get_game_by_client(uid)
{
	for(var i in games){
		var game = games[i];
		if(game.ended)
			continue;
		if(game.clients[uid])
			return game;
	}
	return undefined;
}

function send_lobby_names(){
	// Send to each client
	for(var id in clients){
		// Collect stuff to send
		var dst = []
		for(var id2 in clients){
			// Don't send information of itself
			/*if(id2 == id)
				continue;*/
			var info = {}
			info.id = id2;
			info.name = names[id2];
			info.status = "";
			if(clients_in_game[id2]){
				info.status = "in game";
			}
			dst.push(info);
		}
		client = clients[id];
		console.log("Sending lobby names");
		client.emit('lobby_names', dst);
	}
}

var next_client_uid = 1;

io.sockets.on('connection', function(socket){
	var uid = next_client_uid;
	next_client_uid++;
	console.log("Client connected: uid="+uid);
	clients[uid] = socket;
	names[uid] = "Unknown["+uid+"]";
	
	// Add client automatically to the global game
	//game.add_client(socket);

	socket.emit('assign_id', {id: uid});
	
	socket.on('disconnect', function(data){
		var uid = get_client_uid(socket);
		delete clients[uid];
		send_lobby_names();
		
		var game = get_game_by_client(uid);
		if(game){
			game.players[uid].ended = true;
			if(game.all_ended()){
				game.ended = true;
				delete clients_in_game[uid];
			}

			if(game.all_paths_set()){
				console.log("all paths set in game for uid="+uid+"; emitting all paths.")
				game.emit_all_paths();
			} else {
				console.log("not all paths yet set for game for uid="+uid);
				console.log(game);
			}
		}
	});
	socket.on('update_name', function(data){
		var name = data.name;
		var uid = get_client_uid(socket);
		names[uid] = name;
		console.log("uid="+uid+" reports name "+name);
		send_lobby_names();
	});
	socket.on('request_game', function(data){
		// Create a game from all free players
		var game = new Game();
		console.log("Adding all free players to game");
		for(var uid in clients){
			var socket = clients[uid];
			if(clients_in_game[uid]){
				console.log(uid+" is already in game");
				continue;
			}
			game.add_player(uid);
			clients_in_game[uid] = true;
		}
		game.emit_startup_info();
		games.push(game);
		send_lobby_names();

		setTimeout(function(){
			game.tick();
		}, 1000);
	});
	socket.on('path', function(data){
		var path = data.path;
		var uid = get_client_uid(socket);

		console.log("uid="+uid+" sends path")

		var game = get_game_by_client(uid);
		if(!game){
			console.log("game for uid="+uid+" not found when receiving path");
			return;
		}
		game.players[uid].path = path;
		if(game.all_paths_set()){
			console.log("all paths set in game for uid="+uid+"; emitting all paths.")
			game.emit_all_paths();
		} else {
			console.log("not all paths yet set for game for uid="+uid);
			console.log(game);
		}
	});
	socket.on('ready', function(data){
		var uid = get_client_uid(socket);
		
		var game = get_game_by_client(uid);
		if(game){
			game.players[uid].ended = true;
			delete clients_in_game[uid];
			if(game.all_ended()){
				game.ended = true;
			}
		}
		send_lobby_names();
	});
});

